from __future__ import absolute_import
#
# lutron.system - Lutron interface core
#
# Copyright 2013-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyn
import asyn.utility

import time

from . import core
from . import device

DEBUG = None


#
# A communications channel to forecast.io.
#
class Interface(asyn.Callable, asyn.utility.Idler):
	""" A Lutron system interface.

	"""
	def __init__(self, control, login='nwk', callout=None):
		asyn.Callable.__init__(self, callout=callout)
		asyn.utility.Idler.__init__(self, control)
		self.control = control
		self.login_name = login
		self._io = None


	#
	# Parse out incoming messages, asyn.scan style.
	#
	# The distinction between recoverable and fatal errors is subtle:
	#	foobar:NAK -> couldn't execute the foobar command, retry possible
	#	foobar:NAK [CMD|VAL] -> this is invalid, go away
	#
	SCAN = asyn.scan.Regex([
		(r'login: ', 'login prompt'),
		(r'connection established\r\n', 'login ready'),
		(r'closing connection due to login from (.*?)\r\n', 'login evicted'),
		(r'connection in use\r\n', 'login busy'),
		(r'login incorrect\r\n', 'login error'),
		(r'QSE>', None),										# totally ignore prompts
		(r'^\r\n', None),										# ignore empty lines
		(r'~Error,(\d+)\r\n', 'error'),							# official error indication
		(r'~SYSTEM,(\d+),(.*?)\r\n', 'system'),					# system report
		(r'~DETAILS,(.*?)\r\n', 'details'),						# system discovery item
		(r'~DEVICE,([^,]+),(\d+),(\d+),(.*?)\r\n', 'device'),	# dev, subdev, field, value
		(r'~DEVICE,([^,]+),(\d+),(\d+)\r\n', 'device'),			# dev, subdev, field [no value]
		(r'~MONITORING,.*?\r\n', None),							# ignore monitoring feedback
		(r'~PROGRAMMING,([^,]+),(\d+),([^,]+),(\d+),(.*?)\r\n', 'programming'), # dev, button, dev, subdev, devspec
		(r'~PROGRAMMING,([^,]+),(\d+),3\r\n', 'programming clear'),	# clear program
		(r'~PROGRAMMING,[^,]+,\d+,\d+,.*?\r\n', None),			# programming feedback (ignored)
		('(.*?)\r\n', 'unknown'),
	])

	def open(self, file):
		""" Establish contact with the SAM.

			The file argument must be based on a file descriptor - it can be either
			a genuine File object, or an integer file descriptor. Both sockets and
			local streams are acceptable.

			This method does not start discovery. Call discover() after it if you want it.
		"""
		self._io = self.control.stream(file, callout=self._event)
		self._io.scan = self.SCAN
		self._discovering = False
		self.sysparam = { }
		# waiting for login: prompt to proceed (no retry this stage)


	def device(self, ident):
		""" Find a Lutron device based on a string name. Accepts hex serial or integration id. Case insensitive. """
		try:
			id = int(ident, 16)
			return self.devices.get(id)
		except ValueError:
			pass
		ident = ident.lower()
		for dev in self.devices.values():
			if dev.iid.lower() == ident:
				return dev


	#
	# Close down the object
	#
	def close(self):
		""" Release system resources. """
		self.idle_cancel()
		if self._io:
			self._io.close()
			self._io = None


	#
	# Idle trigger
	#
	def idle(self):
		""" Issue a no-op command to check connectivity. """
		self.command("?SYSTEM,5")

	def idle_timeout(self):
		self.callout("timeout")


	#
	# Discovery.
	# Pacing: We do not use the QSE> prompt (it's too erratic).
	# Instead, we issue SYSTEM commands and sync on the (sequenced) replies.
	#
	def _discover(self):
		self.command("#MONITORING,12,2")		# disable prompts (we don't use them)
		self.command("?DETAILS,ALL_DEVICES")	# what have you got?
		self.command("?SYSTEM,1")				# tell me the time (stage 1 sync marker)
		self.devices = { }						# collecting devices here
		self.nwk = None							# interface device (filled in by its constructor)
		self._discovering = True				# capture all responses during discovery

	def _discover_event(self, ctx, *args):
		""" All incoming messages during discovery are sent here. """
		if ctx.state == 'system':				# SYSTEM,type,stuff
			type = int(args[0])
			if type == 1:						# end of DETAILS
				return self.command("?SYSTEM,2") # tell me the date (stage 2 sync marker)
			elif type == 2:						# end of device discovery
				self._discovering = False
				assert self.nwk					# better have one (or who are we talking to?!)
				for dev in self.devices.values():
					dev._discovery_done()		# chance to respond to circular dependencies
				return self.callout('discover', 'done')
		elif ctx.state == 'details':
			return self._discover_device(*args)
		elif ctx.state == 'device':
			return self._device_update(*args)
		elif ctx.state == 'programming':
			return self._program_update(*args)
		self.callout('UNEXPECTED', 'discovery', ctx, *args)

	def _discover_device(self, form):
		desc = { }
		for arg in form.split(','):
			key, _, value = arg.partition(':')
			desc[key] = value
		dev = device.makedev(self, desc)
		self.devices[dev.serial] = dev
		self.callout('discover', 'device', dev)


	#
	# Omnibus event handler.
	# Dispatches events to the system level if we can figure out where it belongs.
	# Everything global gets called out.
	#
	def _event(self, ctx, *args):
		if ctx.error:
			return self.callout(ctx)
		if ctx.state == 'CLOSE':			# connection lost
			self._io = None					# any retry to be done by upstream
			return self.callout('close')	# ... so let them know
		if ctx.state == 'END':
			return self.callout(ctx)
		if DEBUG: DEBUG('<-', ctx, *args)
		if self._discovering:				# active discovery eats all events
			return self._discover_event(ctx, *args)
		if ctx.state == 'login prompt':		# initial login prompt (send login name)
			self.command(self.login_name)
		elif ctx.state == 'login ready':	# login complete; start discovery
			self.callout('login', 'ready')
			self._discover()
		elif ctx.state == 'login error':	# login rejected; give up
			self._io.close()
			self._io = None
			self.callout('login', 'error')
		elif ctx.state == 'login busy':		# login busy; give up but caller may try again
			self._io.close()
			self._io = None
			self.callout('login', 'busy')
		elif ctx.state == 'login evicted':	# we just got kicked out by another connection
			self._io.close()
			self._io = None
			self.callout('login', 'evicted', *args)
		elif ctx.state == 'device':			# (iid, component#, action# [, stuff])
			self._device_update(*args)
		elif ctx.state == 'programming':
			self._program_update(*args)
		elif ctx.state == 'programming clear':
			self._program_clear(*args)
		elif ctx.state == 'system':			# response to SYSTEM,n requests (system info)
			self.sysparam[args[0]] = args[1]
		elif ctx.state == 'unknown':		# something else
			self.callout('UNEXPECTED', 'unknown input', ctx, *args)
		else:
			self.callout('UNEXPECTED', 'event type', ctx, *args)
		self.idle_activity()

	def _device_update(self, dev, comp, action, *args):
		dev = self.device(dev)
		comp = int(comp)
		action = int(action)
		if dev:
			dev.update(comp, action, *args)
		else:
			self.callout('UNEXPECTED', 'device update', dev, comp, action, *args)

	def _program_update(self, nwk, button, dev, subdev, conf):
		nwk = self.device(nwk)
		lift, tilt = ([int(s) if s else None for s in conf.split(',')] + [None])[0:2]
		nwk.updateProgram(int(button), dev, int(subdev), lift, tilt)

	def _program_clear(self, nwk, button):
		nwk = self.device(nwk)
		nwk.clearProgram(int(button))


	#
	# Command processing
	#
	def command(self, *args):
		""" Send a command to the SAM. """
		cmd = ','.join(map(str, args))
		if DEBUG: DEBUG('->', cmd)
		self._io.write_a(f'{cmd}\r\n')

	def __repr__(self):
		return "<NWK@%r>" % self._io
