#encoding: utf-8
#
# Lutron system interface for Indigo 6+
#
# Copyright 2013-2017 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import time

import lutron

import asyn
import cyin.devstate
from cyin import iom, plug
from cyin import log, debug, error
from cyin.asynplugin import action
from cyin.check import *


#
# The plugin represents tilts as -100..100. Zero is flat (parallel to ground).
# Positive numbers tilt "normal" (angled upwards looking out).
# Negative numbers tilt "counter" (looking down sighting along the slats).
# In Lutron speak, 0 is full normal, 50 is flat, and 100 is full counter.
#
def tilt_i2l(tilt):
	""" Map Indigo tilt (-100..100) to Lutron tilt (0..100). """
	if tilt is not None:
		return 50 + tilt/2

def tilt_l2i(tilt):
	""" Map Lutron tilt (0..100) to Indigo tilt (-100..100). """
	if tilt is not None:
		return 2*tilt - 100


#
# A Lutron QSE NWK interface device.
#
class Interface(cyin.devstate.IPDevice, cyin.devstate.Idler):
	""" A Lutron NWK or similar interface. This is what we talk to.

		This represents the Indigo-facing side of the Lutron installation.
	"""
	login = cyin.PluginProperty(type=str)

	@property
	def nwk(self): return self.target

	DEFAULT_PORT = lutron.PORT
	SERIAL_CONFIG = dict(baudrate=9600,)

	def connected(self, socket):
		""" IPDevice hook: network connection established. """
		self.target = lutron.Interface(cyin.plugin, login=self.login, callout=self._event)
		self.proceed("exploring")
		self.nwk.open(socket)
		self.idle_update()

	@action
	def raw_command(self, action):
		""" Issue an arbitrary low-level Lutron command to the interface. """
		try:
			self.target.command(action.command)
		except Exception as e:
			error(self.name, 'cannot send command "%s": %s' % (action.command, e))

	def _event(self, ctx, *args):
		""" Handle events sent from the Lutron interface layer. """
		if ctx.error:
			return self.fail_soft(ctx)
		if ctx.state == 'discover':
			type = args[0]
			if type == 'device':
				dev = args[1]
				debug('discovered %s (%s)' % (dev.description, dev.NAME))
			elif type == 'done':
				debug('discovery complete')
				self.proceed("ready", recovered=True)
		elif ctx.state == 'error':
			error("error:", *args)
		elif ctx.state == 'END':
			self.reconnect()
		elif ctx.state == 'close':
			pass
		elif ctx.state == 'timeout':
			self.fail_soft("no response from device")
		elif ctx.state == 'login':
			what = args[0]
			if what == 'ready':
				debug(self.name, "user", self.login, "logged in")
			elif what == 'error':
				self.fail_hard("login %s rejected" % self.login)
				self.target.close()
			elif what == 'evicted':
				self.fail_hard("bumped from address %s" % args[1])
			elif what == 'busy':
				self.fail_soft("login busy")
			else:
				error("unexpected login event", ctx, *args)
		else:
			error("unexpected lutron event", ctx, *args)


#
# The base class of all Lutron device-based devices.
# Subclasses of this are reported out of device discovery.
# There are also duck-typed "devices" that are not subclasses of this.
#
class Device(cyin.devstate.SubDevice):
	""" An Indigo device based on a Lutron device attached to a control interface.

		This has no behavior; that's specific to the kind of device.
	"""
	def set_subaddress(self, subaddr):
		self.serial = int(subaddr)

	@property
	def device(self): return self.target

	def setup(self, ctx=None):
		self.nwk = self.hostdev.nwk
		if self.serial not in self.nwk.devices:
			return self.fail_hard("no %s 0x%x in %s" % (self.PARTNAME, self.serial, self.hostdev.name))
		self.target = self.nwk.devices[self.serial]
		self.set_display_address(self.target.description)
		self.device.add_callout(self._event)
		self._updateState()
		self.proceed("ready")


#
# Common base class of shade devices.
#
class Shades(Device):
	""" A Lutron shade device (roller shades or venetian blinds).

		This responds to dimmer and relay standard actions.
	"""
	onOff = cyin.DeviceState(type=bool, name="onOffState")
	brightnessLevel = cyin.DeviceState(type=int)
	motion = cyin.DeviceState(type=int)

	PARTNAME = 'shade'

	@action
	def standard_status(self):
		self.device.refresh()

	def _event(self, ctx, *args):
		if ctx.error:
			return self.fail_soft(ctx)
		if ctx.state in ['value', 'tilt']:
			self._updateState()
		elif ctx.state == 'motion':
			self.motion, = args
		else:
			error("unexpected shades device event", ctx, *args)


#
# Simple roller shade devices
#
class RollerShades(Shades):
	""" A (single) roller shade or drape device.

		RollerShades have only one degree of freedom, and map perfectly onto
		Indigo dimmer devices.
	"""
	PARTNAME = "roller shade"

	def _updateState(self):
		self.brightnessLevel = self.device.value
		self.onOff = self.brightnessLevel > 0

	@action
	def move(self, action):
		self.device.moveTo(action.position)

	@action
	def stop_motion(self, action):
		self.device.move(lutron.SHADE_STOP)

	@action
	def standard_switch(self, active):
		self.device.moveTo(100 if active else 0)

	@action
	def standard_brightness(self, intensity):
		self.device.moveTo(intensity)


#
# A mix-in providing standard dimmer/relay event behavior for venetian blinds.
#
class VenetianBehavior(object, metaclass=cyin.iom.IOMetaMap, register=False):
	""" A mix-in class providing device response to standard dimmer/relay handling.

		Dimmer/relay commands move one axis (lift or tilt), depending on the current
		setting of the dim_tilt property. The other axis is either left alone,
		or is set to a default value depending on the default_tilt/default_position
		properties.
		If we control tilt, it will range [0..100] in Indigo coordinates, and the
		tilt_negative property will flip this to range [0..-100] instead, effectively
		reversing the tilt direction. Either way, you only get half of the movement
		range.

		Subclass/side class must provide self.device pointing to an object
		responding to the moveTo method.

		This class does not respond to custom movement actions.
	"""
	dim_tilt = cyin.PluginProperty(type=bool)
	default_tilt = cyin.PluginProperty(type=int, required=False, eval=True,
		check=[check_range(min=-100, max=100)])
	tilt_negative = cyin.PluginProperty(type=bool)
	default_position = cyin.PluginProperty(type=int, required=False, eval=True,
		check=[check_range(min=0, max=100)])
		
	tilt = cyin.DeviceState(type=int)


	def standardMove(self, value):
		""" Execute a standard switch/dimmer command. """
		debug("standard_move to", value, "dim_tilt", self.dim_tilt, "negative", self.tilt_negative)
		if self.dim_tilt:		# set tilt, default position
			if self.tilt_negative:
				value = -value
			debug("sending tilt", value, "->", tilt_i2l(value))
			self.device.moveTo(self.default_position, tilt_i2l(value))
		else:					# set position, default tilt
			self.device.moveTo(value, tilt_i2l(self.default_tilt))

	def _updateState(self):
		self.tilt = tilt_l2i(self.device.tilt)
		debug("update_state tilt", self.device.tilt, "->", self.tilt)
		if self.dim_tilt:
			value = self.tilt
			if self.tilt_negative:
				value = -value
			self.brightnessLevel = value
			debug("set brightnessLevel to", self.brightnessLevel)
		else:
			self.brightnessLevel = self.device.value
		self.onOff = self.brightnessLevel > 0

	@action
	def standard_switch(self, active):
		self.standardMove(100 if active else 0)

	@action
	def standard_brightness(self, intensity):
		self.standardMove(intensity)



#
# Venetian Blinds devices
#
class VenetianBlinds(VenetianBehavior, Shades):
	""" A (single) Venetian Blind device. It moves and tilts.

		VenetianBlinds execute custom Move commands, setting either lift or tilt
		or both as requested.
		VenetianBlinds also obeys standard switch/dimmer commands. The linkage
		can be to either lift or tilt, with the other optionally defaulted.
		In other words, normal dimmer commands can be used to switch either aspect
		of the blinds, depending on device configuration.
	"""
	PARTNAME = "blind"

	@action
	def move(self, action):
		debug("direct move position", action.position, "tilt", action.tilt, "->", tilt_i2l(action.tilt))
		self.device.moveTo(action.position, tilt_i2l(action.tilt))

	@action
	def stop_motion(self, action):
		self.device.move(lutron.SHADE_STOP)


#
# A "phantom button" of a QSE interface device
#
class Phantom(VenetianBehavior, cyin.devstate.SubDevice):
	""" A Phantom Button on the primary NWK interface. Buttons also act as device groups. """

	onOff = cyin.DeviceState(type=bool, name="onOffState")
	brightnessLevel = cyin.DeviceState(type=int)
	motion = cyin.DeviceState(type=int)

	PARTNAME = 'button'
	
	config_version = 2

	def set_subaddress(self, subaddr):
		self.bno = int(subaddr)

	@property
	def device(self): return self.button	# for VenetianBehavior[moveTo]

	def setup(self, ctx=None):
		self.nwk = self.hostdev.nwk
		if self.bno not in self.nwk.nwk.phantoms:
			return self.fail_hard("no button %d in %s" % (self.serial, self.hostdev.name))
		self.button = self.nwk.nwk.phantoms[self.bno]
		self.button.add_callout(self._event)
		self._updateState()
		self.motion = self.button.motion
		self.proceed("ready")

	@action
	def do_phantom(self, action):
		self.button.press()

	@action
	def move(self, action):
		self.button.moveTo(action.position, tilt_i2l(action.tilt))

	def _updateState(self):
		self.brightnessLevel = self.device.value
		self.onOff = self.brightnessLevel > 0
		self.tilt = tilt_l2i(self.device.tilt)

	def _event(self, ctx, *args):
		if ctx.error:
			return self.fail_soft(ctx)
		if ctx.state in ['value', 'tilt']:
			self._updateState()
		elif ctx.state == 'motion':
			self.motion, = args
		else:
			error("unexpected button event", ctx, *args)


#
# Actions
#
class Move(cyin.DeviceAction):
	position = cyin.PluginProperty(type=int, required=False, eval=True, check=[check_range(min=0, max=100)])
	tilt = cyin.PluginProperty(type=int, required=False, eval=True, check=[check_range(min=-100, max=100)])

	class UI(cyin.ConfigUI):
		def description(self):
			if self.position and self.tilt:
				return "move to %s, tilt to %s" % (self.position, self.tilt)
			elif self.position:
				return "move to %s" % (self.position, )
			elif self.tilt:
				return "tilt to %s" % (self.tilt, )

class StopMotion(cyin.DeviceAction):
	pass

class DoPhantom(cyin.DeviceAction):
	pass

class Raw_Command(cyin.DeviceAction):
	command = cyin.PluginProperty(type=str, eval=True)
	class UI(cyin.ConfigUI):
		def description(self):
			return 'send %s to %s' % (self.command, self.device.name)


#
# Menu filters
#
class LutronDeviceFilter(cyin.MenuGenerator):
	""" Menu filter - Lutron devices by (lutron) device type name. """
	def generate(self):
		cls = lutron.DEVICETYPES[self.filter]
		for interface in Interface.all():
			if interface.ready():
				for dev in interface.nwk.devices.values():
					if isinstance(dev, cls):
						yield (
							"%d@%s" % (interface.id, dev.serial),
							"%s @ %s (%s)" % (interface.name, dev.description, dev.NAME)
						)

class LutronPhantomFilter(cyin.MenuGenerator):
	""" Menu filter - programmed lutron phantom buttons. """
	def generate(self):
		for interface in Interface.all():
			for button in interface.target.nwk.phantoms.values():
				yield (
					"%d@%d" % (interface.id, button.bn),
					"%s @ button %d" % (interface.name, button.bn)
				)


#
# The plugin. Nothing to see here. Move along.
#
class Plugin(cyin.asynplugin.Plugin):
	pass
